import os

import matplotlib.pyplot as plt

import numpy as np

import time

import skimage
from skimage import io, util
from skimage.morphology import disk
from skimage.filters import threshold_sauvola
from skimage.draw import ellipse_perimeter

from scipy import stats, integrate
from scipy.signal import butter, filtfilt

import pandas as pd


def average_list(generic_list):
    """ function that returns the average value of all values in list """
    return sum(generic_list) / len(generic_list)


def normalize_list(generic_list):
    """ function that returns list with values from 0-1 according to max and/or min values of input """
    max_or_min = max([abs(item) for item in generic_list])
    return [item / max_or_min for item in generic_list]


def normalize_array(gen_array):
    """ function that returns numpy array with values from 0-1 according to max and/or min
    values of input """
    max_gen_array = np.amax(gen_array)
    min_gen_array = np.amin(gen_array)
    for item_x in range(0, len(gen_array[:, 0])):
        for item_y in range(0, len(gen_array[0, :])):
            gen_array[item_x, item_y] = (gen_array[item_x, item_y] - min_gen_array) / (max_gen_array - min_gen_array)
    return gen_array


def get_x_pixel_values(generic_image):
    """ function that returns a normalised list of the pixel values from slices along the x axis """
    return normalize_list([sum(generic_image[:, column_x]) for column_x in range(0, len(generic_image[0, :]))])


def get_y_pixel_values(generic_image):
    """ function that returns a normalised list of the pixel values from slices along the y axis """
    return normalize_list([sum(generic_image[column_y, :]) for column_y in range(0, len(generic_image))])


def crop_image(generic_image, lim_x, lim_y):
    """ function that returns a cropped version of the input image """
    # generic_image = generic image
    # lim_x = integer tuple with limits along x
    # lim_y = integer tuple with limits along y
    return generic_image[lim_y[0]:lim_y[1], lim_x[0]:lim_x[1]]


def find_substrate_line_index(y_values, y_values_derivative):
    """ function that returns the y coordinate at which the interface between the substrate and the
        atmosphere is located """
    # y_values = list with the normalised sum of all pixel values along the rows in the y direction. Should
    #            be the output of get_y_pixel_values
    # y_values_derivative = derivative of the y_values

    substrate_found = False
    while substrate_found == False:
        substrate_line_index = int(
            np.argmin(y_values_derivative)
        )
        # checks if the initial guess is not out-of-bounds (too small) for considered image
        if substrate_line_index - len_image_y * 0.1 < 0:
            dummy_value = 0
        else:
            dummy_value = int(round(substrate_line_index - len_image_y * 0.1))
        # checks if initial guess is not out-of-bounds (too large)
        if substrate_line_index + int(round(len_image_y * 0.15, 0)) > len_image_y:
            check_2 = len_image_y - 1
        else:
            check_2 = substrate_line_index + int(round(len_image_y * 0.15, 0))

        # makes sure the smallest derivative is not located near the top of the bead
        if y_values[dummy_value] <= 0.9 and y_values[check_2] < 0.15:
            substrate_found = True
        else:
            # substrate not found, being a "false" positive, so the value is zero-ed out
            y_values_derivative[substrate_line_index] = 0

    return substrate_line_index


def check_for_nonzero(a, x, y, p_d):
    """ function that finds "gaps" in the denoised sauvola threshold image. This is needed before the
        binarisation stage, as it allow the gaps to be filled. If this step is not done,
         vertical gaps will appear in the binarised image, compromising the derivatives of the normalised sums
        """
    # a = sauvola thresholded image
    # x = x coordinate
    # y = y coordinate
    len_x = len(a[0, :])
    len_y = len(a)

    if x > len_x - 1:
        x = len_x - 1
    if x < - len_x:
        x = -len_x + 1
    if y > len_y - 1:
        y = len_y - 1
    if y < - len_y:
        y = -len_y + 1

    return_value = False
    if a[y, x] == True and x < len_x - 1:
        if (y, x) not in p_d:
            return_value = True
    return return_value, y, x


def get_search_indices(r, y, x):
    x_ind = []
    y_ind = []
    len_x = 4 * r
    for pix in range(0, r + 1):
        x_ind.append(x + pix)
        y_ind.append(y - r)

    for pix in range(r + 1, 3 * r + 1):
        x_ind.append(x + r)
        y_ind.append(y_ind[-1] + 1)

    return y_ind, x_ind


def evaluate_points(dis, y, x):
    metric = []

    for d in dis:
        if d == 0:
            metric.append(0)
        else:
            metric.append(1 / d)

    ind = np.argmax(metric)

    return y[ind], x[ind]


def nearest_nonzero_idx(a, x, y, top_pixel_x, pixel_dictionary):
    current_pixel_x = x
    current_pixel_y = y
    radius = 0
    found_points_x = []
    found_points_y = []
    found_dist = []
    test_pixel = False
    certainty = 0
    while len(found_dist) < 10:
        radius += 1
        y_indices, x_indices = get_search_indices(radius, current_pixel_y, current_pixel_x)
        for pixel_y, pixel_x in zip(y_indices, x_indices):
            test_pixel, current_pixel_y, current_pixel_x = check_for_nonzero(a, pixel_x, pixel_y, pixel_dictionary)
            if test_pixel == True:
                found_points_x.append(current_pixel_x)
                found_points_y.append(current_pixel_y)
                if x - current_pixel_x < 0:
                    found_dist.append(
                        -np.sqrt(
                            (x - current_pixel_x) ** 2 +
                            (y - current_pixel_y) ** 2
                        )
                    )
                else:
                    found_dist.append(
                        np.sqrt(
                            (x - current_pixel_x) ** 2 +
                            (y - current_pixel_y) ** 2
                        )
                    )
    current_pixel_y, current_pixel_x = evaluate_points(found_dist, found_points_y, found_points_x)

    return int(round(current_pixel_x, 0)), int(round(current_pixel_y, 0))


def get_bead_limits(image, image_without_substrate, x_values, x_values_derivative, integration_value,
                    inverted_x_values):
    stop_criteria = 0.98

    midway = max(inverted_x_values)
    midway_index = np.argmax(inverted_x_values)
    left_l = int(round(midway_index * 0.9, 0))
    right_l = int(round(midway_index * 1.1, 0))
    int_val = integrate.simpson(inverted_x_values[left_l:right_l])
    while int_val < stop_criteria * integration_value and left_l > 0 and right_l < len(image[0, :]) - 1:
        left_in = integrate.simpson(inverted_x_values[:left_l])
        right_in = integrate.simpson(inverted_x_values[right_l:])
        if left_in > right_in:
            left_l -= 1
        else:
            right_l += 1
        int_val = integrate.simpson(inverted_x_values[left_l:right_l])
    return left_l, right_l


def find_zero_columns(image):
    return_value = []
    len_x = len(image[0, :])
    for pixel_x in range(0, len_x):
        dummy_column = image[:, pixel_x]
        if True not in dummy_column:
            return_value.append(pixel_x)
    return return_value


def find_non_consecutive_pixels(l):
    ind = 0
    dum = 0
    non_cons = []
    while ind < len(l) - 2:
        ind += 1
        if l[ind - 1] + 1 != l[ind]:
            if l[ind] != l[ind + 1]:

                non_cons.append((l[dum], l[ind - 1]))
                dum = ind
            else:
                dum = l[ind - 1]

    return non_cons


def return_first_true(co):
    rv = 0
    for ind in range(0, len(co)):
        if co[ind] == True:
            rv = ind
            break
    return rv


def fill_voids(im, non_cons):
    """ function that fills gaps in sauvola threshold image """
    # im = binarised image
    for pc in non_cons:
        # find the minimum y for which pixels exist
        if pc[0] == 0:
            top_left = return_first_true(im[:, non_cons[1][1] + 1])
            top_right = top_left
        else:
            top_left = return_first_true(im[:, pc[0] - 1])
            top_right = return_first_true(im[:, pc[1] + 1])
        rr, cc = skimage.draw.line(top_left, pc[0], top_right, pc[1])
        im[rr, cc] = 1
    if len(non_cons) > 1:
        if non_cons[-1][1] + 1 < len(im[0, :]):
            rr, cc = skimage.draw.line(top_right, non_cons[-1][1], top_right, len(im[0, :]) - 1)
            im[rr, cc] = 1

    return im


def low_pass_filter(x_data, y_data, load_frequency_hz=1, order=2, sampling_frequency=500):
    """ function that returns a """
    signal_period = 1 / load_frequency_hz

    # Filter Parameters
    cycle_period = 1
    total_samples = sampling_frequency * cycle_period
    signal_frequency = load_frequency_hz

    cutoff = load_frequency_hz * 10

    nyquist_frequency = 0.5 * sampling_frequency

    n = int(cycle_period * sampling_frequency)
    normal_cutoff = cutoff / nyquist_frequency
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return filtfilt(b, a, y_data)


def draw_line_from_slope(im, y1, x1, m, val=0):
    """ function that draws a line through initial x and y coordinates and the respective line's slope,
        in the input image"""
    # y1 =
    # y2 =
    # finding b
    b = y1 - m * x1
    # finding x_p_2
    x_p_2 = - int(round(b / m, 0))
    y_p_2 = 0

    if x_p_2 > len(im[0, :]) - 1:
        x_p_2 = len(im[0, :]) - 1
        y_p_2 = int(round(m * x_p_2 + b, 0))

    if x_p_2 < 0:
        x_p_2 = 0
        y_p_2 = int(round(b, 0))

    # *(x1, x2) (y1, y2)
    # rr, cc = skimage.draw.line(current_pixel_y, current_pixel_x, next_pixel_y, next_pixel_x)
    rr, cc = skimage.draw.line(y1, x1, y_p_2, x_p_2)
    im[rr, cc] = val


def get_line_bead_maximum(bi):
    """ Returns (pixel y, pixel x) pertaining to the
        point where the line bead is the tallest,
        in relation to the substrate
    """
    # input variables:
    # bi - binerized image of the line bead

    return_y = 0
    return_x = 0

    while bi[return_y, return_x] == 1 and return_y < len(bi[:, 0]) - 1:
        return_x += 1
        if return_x == len(bi[0, :]) - 1:
            return_y += 1
            return_x = 0


    return (return_y, return_x)


def get_height(m, s, conv_f):
    """ function that returns the line bead height from """
    return (s - m) * conv_f


def get_width(ll, rl, conv_f):
    """ function that returns the line bead width """
    return (rl - ll) * conv_f


def get_left_wettability_angle(bi, left_l, substrate_line, height, criteria, dy=1):
    """ function that returns the wettability angle of the left side of the line bead """
    # input variables:
    # im = binarized image with only the top part of the line bead
    # left_l = point at which x values start being analysed
    # substrate_line = index where substrate meets bead, also the cut-off of bi
    # criteria = the fraction of the height that will be considered when performing a linear regression that
    #            returns the slope of all points that shape the bead
    # dy [optional] = interval between points considered for the left_y_list list

    # list that will contain x values pertaining to points of the line bead
    left_x_list = []

    y_top_limit = int(round(substrate_line-5 - criteria * height, 0))
    # list that contains y values pertaining to the points of the line bead
    left_y_list = np.arange(substrate_line - 5, y_top_limit, -dy)

    for pixel_y in left_y_list:
        pixel_x = 1
        while bi[pixel_y, left_l + pixel_x] == 1 and left_l + pixel_x < len(bi[0, :]) - 1:
            pixel_x += 1
        left_x_list.append(left_l + pixel_x)

    # perform linear regression with x and y values
    left_wettability_regression = stats.linregress(left_x_list, left_y_list)

    left_slo, left_ang = get_angle_from_regression(left_wettability_regression)
    # adapts angle to correct value (due to arctan)
    if left_ang < 0:
        left_ang += 180

    return left_slo, left_ang


def get_right_wettability_angle(bi, right_l, substrate_line, height, criteria, dy=1):
    """ function that returns the wettability angle of the right side of the line bead """
    # input variables:
    # im = binarized image with only the top part of the line bead
    # right_l = point at which x values start being analysed
    # substrate_line = index where substrate meets bead, also the cut-off of bi
    # criteria = the fraction of the height that will be considered when performing a linear regression that
    #            returns the slope of all points that shape the bead
    # dy [optional] = interval between points considered for the left_y_list list

    # list that will contain x values pertaining to points of the line bead
    right_x_start = min(right_l, len(bi[0, :]) - 15)

    right_x_list = []

    # selects appropriate y coordinate from which the contour will be found
    start_pixel_y = substrate_line - 1
    while bi[start_pixel_y, right_x_start] == 0 and bi[start_pixel_y, right_x_start - 1] == 0:
        start_pixel_y -= 1

    # maximum coordinate that will be considered when accounting points in the contour
    # the smaller the value, the less height will be accounted for when computing the angle
    y_top_limit = int(round(start_pixel_y - height * criteria, 0))

    right_y_list = np.arange(start_pixel_y,  y_top_limit, - dy)

    # finds all occurances in the contour of the bead between the start and the y_top_limit
    for pixel_y in right_y_list:
        pixel_x = -1
        while bi[pixel_y, right_x_start + pixel_x] == 1 and right_x_start + pixel_x > 0:
            pixel_x -= 1
        right_x_list.append(right_x_start + pixel_x)

    # performs linear regression with selected points
    right_wettability_regression = stats.linregress(right_x_list, right_y_list)
    # gets angle and slope from the linear regression
    right_slo, right_ang = get_angle_from_regression(right_wettability_regression)
    # converts value (due to arctan)
    if right_ang > 0:
        right_ang -= 180

    return right_slo, -right_ang



def get_angle_from_regression(regr):
    # if line angle is vertical, this line avoids division by 0
    if regr.rvalue == 0:
        slo = 10 ** 6

        wett_angle = 90
    # else wettabilty angle is computed regularly
    else:
        slo = regr.slope
        wett_angle = -np.arctan(regr.slope) * 180 / np.pi

    return slo, wett_angle



def edge_detection_thru_sauvola_edge(im_gr):
    l_x = len(im_gr[0, :])
    l_y = len(im_gr[:, 0])
    thresh_sauvola = threshold_sauvola(im_gr)
    binary_sauvola = im_gr < thresh_sauvola

    # apply edge filter to further image detection
    image_edge = skimage.filters.scharr(im_gr)
    image_edge = normalize_array(image_edge)

    # filter image edge detection
    binary_edge = np.full((l_y, l_x), False)
    for pixel_x in range(0, l_x):
        for pixel_y in range(0, l_y):
            if image_edge[pixel_y, pixel_x] > 0.1:
                binary_edge[pixel_y, pixel_x] = True

    # perform OR boolean operation on both edge detection procedures
    return binary_sauvola | binary_edge


def close_gaps_in_threshold(thr):
    non_false_pixels_x = np.where(thr == True)[1]
    non_false_pixels_y = np.where(thr == True)[0]

    non_false_pixels_x_sort_order = np.argsort(non_false_pixels_x)
    non_false_pixels_sorted_x_x = np.array(non_false_pixels_x)[non_false_pixels_x_sort_order]
    non_false_pixels_sorted_x_y = np.array(non_false_pixels_y)[non_false_pixels_x_sort_order]

    current_pixel_x = non_false_pixels_sorted_x_x[0]
    current_pixel_y = non_false_pixels_sorted_x_y[0]


    pixel_encyclopedia = []
    pixel_encyclopedia.append((current_pixel_y, current_pixel_x))

    zero_c = find_zero_columns(thr)

    if zero_c:
        zero_d = find_non_consecutive_pixels(zero_c)
        thr = fill_voids(thr, zero_d)

    return thr

def binarise_bead(len_x, len_y, srd):
    blank = np.zeros((len_y, len_x))
    blank_2 = np.zeros((len_y, len_x))
    for pixel_x in range(0, len_x):
        pixel_y = 0
        while srd[pixel_y, pixel_x] == False and pixel_y < len_y - 1 and \
                srd[pixel_y, pixel_x-1] == 0:
            blank[pixel_y, pixel_x] = 1
            blank_2[pixel_y, pixel_x] = 1
            pixel_y += 1


    for pixel_y in range(0, len_y):
        pixel_x = 0
        while srd[pixel_y, pixel_x] == False and pixel_x < len_x - 1:
            blank[pixel_y, pixel_x] = 1
            pixel_x += 1

        pixel_x = len_x - 1
        while srd[pixel_y, pixel_x] == False and pixel_x > 0:
            blank[pixel_y, pixel_x] = 1
            pixel_x -= 1


    pixel_y = 0
    pixel_y_2 = -1
    while pixel_y_2 < len_y - 1:
        pixel_y_2 += 1
        pixel_y = pixel_y_2
        pixel_x = len_x - 1
        while pixel_x > 0 and pixel_y < len_y -1 and srd[pixel_y, pixel_x] == 0:
            blank[pixel_y, pixel_x] = 1
            pixel_x -= 1
            pixel_y += 1

    pixel_y = 0
    pixel_y_2 = -1
    while pixel_y_2 < len_y - 1:
        pixel_y_2 += 1
        pixel_y = pixel_y_2
        pixel_x = 0
        while pixel_x < len_x - 1 and pixel_y < len_y - 1 and srd[pixel_y, pixel_x] == 0:
            blank[pixel_y, pixel_x] = 1
            pixel_x += 1
            pixel_y += 1

    return blank, blank_2


def print_log(*args, **kwargs):
    print(*args)
    print(*args, **kwargs)





counter = 0

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__))
)

notepad_file = os.path.join(__location__, 'criteria.txt')


with open(notepad_file) as opened_file:
    fo = opened_file.readlines()[0:9]

    # pixel to millimeter conversion factor
    px_2_mm = fo[0]
    px_2_mm = float(px_2_mm.split('#')[0])

    # whether image is inverted or not
    invert_image = fo[1]
    invert_image = int(invert_image.split('#')[0])

    # fraction of the height considered to measure wettability angles
    fraction_height_wettability = fo[2]
    fraction_height_wettability = float(fraction_height_wettability.split('#')[0])

    # whether or not the images with the determined properties will be saved or not
    save_images = fo[3]
    save_images = int(save_images.split('#')[0])

    # whether or not the images and some key plots will be saved or not
    save_image_collection = fo[4]
    save_image_collection = int(save_image_collection.split('#')[0])

    # whether a dataframe with the computed variables will be saved
    save_data = fo[5]
    save_data = int(save_data.split('#')[0])

    # directory to which the outputs will be saved
    save_directory = fo[6]
    save_directory = save_directory.split(' ')
    print(save_directory)
    dum = 0
    while save_directory[dum] != '#':
        dum += 1
    dum -= -1
    while save_directory[dum] == ' ':
        del save_directory[dum]
        dum -= 1
    save_directory = ' '.join(save_directory)
    save_directory = save_directory.split('#')[0].replace('\\', '/')
    save_directory = os.path.abspath(save_directory)

    # directory to which images are searched
    search_directory = fo[7]
    search_directory = search_directory.split(' ')
    print(search_directory)
    dum = 0
    while search_directory[dum] != '#':
        dum += 1
    dum -= -1
    while search_directory[dum] == ' ':
        del search_directory[dum]
        dum -= 1
    search_directory = ' '.join(search_directory)
    search_directory = search_directory.split('#')[0].replace('\\', '/')
    search_directory = os.path.abspath(search_directory)

    # acceptable filetypes
    save_image_collection = fo[8]
    save_image_collection = save_image_collection.split('#')[0]
    save_image_collection = save_image_collection.replace(' ', '')
    save_image_collection = save_image_collection.split(',')


image_list = []

for root, dir, files in os.walk(search_directory, topdown=True):
    dir.clear()
    for name in files:
        for type_of_file in save_image_collection:
            if name.endswith(type_of_file):
                image_list.append(os.path.join(root, name))

start_time = time.time()


text_directory = os.path.join(save_directory, 'Run-log.txt')

with open(text_directory, 'w') as text_file:
    print_log('+++++ THANK YOU FOR USING THIS CODE +++++', file=text_file)
    print_log('Code written, developed and tested by:', file=text_file)
    print_log(' Jorge Gil;            ORCID: 0000-0002-5225-8084;  email: jgil@inegi.up.pt', file=text_file)
    print_log(' João Tavares, PhD;    ORCID: 0000-0001-7603-6526;', file=text_file)
    print_log(' Abílio de Jesus, PhD; ORCID: 0000-0002-1059-715X', file=text_file)
    print_log(' Beatriz Silva, PhD;   ORCID: 0000-0002-5284-8391', file=text_file)
    print_log(' Fátima Vaz, PhD;      ORCID: 0000-0003-1629-523X', file=text_file)
    print_log(' Ana Reis, PhD;        ORCID: 0000-0003-0256-1488', file=text_file)
    print_log(f'Images will be fetched in the following directory: {search_directory}', file=text_file)

    print_log(f'Number of found images: {len(image_list)};', file=text_file)
    print_log(f'Conversion factor between pixel to mm: {px_2_mm};', file=text_file)
    if save_images == 1:
        print_log(f'Images with visual representation of the properties will be saved;', file=text_file)
    else:
        print_log(f'Images with visual representation of the properties will not be saved;', file=text_file)

    if invert_image == 1:
        print_log(f'Image color scheme will be inverted;', file=text_file)
    else:
        print_log(f'Image color scheme will not be inverted', file=text_file)

    if save_image_collection == 1:
        print_log(f'Images with data and plots will be saved;', file=text_file)
    else:
        print_log(f'Images with data and plots will not be saved;', file=text_file)

    if save_data == 1:
        print_log('Property database will be saved;', file=text_file)
    else:
        print_log('Property database will not be saved;', file=text_file)

    if save_images == 1 or save_image_collection == 1 or save_directory == 1:
        print_log(f'Outputs will be saved to: {save_directory}', file=text_file)

line_bead_name_list = []
height_list = []
width_list = []
penetration_list = []
area_above_substrate_list = []
area_below_substrate_list = []
left_wettability_angle_list = []
right_wettability_angle_list = []
dilution_proportion_list = []



counter = 0
for path in image_list:

    # counts time
    start_cycle_time = time.time()

    # reads image
    image_gray = io.imread(path, as_gray=True)

    image_original = image_gray.copy()

    if invert_image == 1:
        image_original = util.invert(image_original)

    current_bead = path.split('/')[-1]
    current_bead = current_bead.split('.')[0]
    with open(text_directory, 'a') as text_file:
        print_log('+++++++++++++++++++++++++++++++++++++++++++', file=text_file)
        print_log(f'{counter} images analysed; \n   {len(image_list) - counter} images to go;', file=text_file)
        print_log(f'    Current bead: {current_bead}', file=text_file)
    line_bead_name_list.append(current_bead)


    # acquire image lengths
    len_image_x = len(image_gray[0, :])
    len_image_y = len(image_gray)


    """ Finding the substrate """
    # apply sauvola threshold to detect image edges
    sauvola_result = edge_detection_thru_sauvola_edge(image_gray)

    sauvola_result_original = sauvola_result.copy()

    # denoise the image through white top hat, with disk of 1 pixel radius
    sauvola_result_noise = sauvola_result.copy()
    sauvola_result_noise = skimage.morphology.white_tophat(sauvola_result_noise, disk(1))
    sauvola_result_denoised = sauvola_result.copy()
    sauvola_result_denoised[:15, :] = False


    # removing noise
    for pixel_x in range(0, len_image_x):
        for pixel_y in range(0, len_image_y):
            if sauvola_result_noise[pixel_y, pixel_x] == True:
                sauvola_result_denoised[pixel_y, pixel_x] = False


    sauvola_result_denoised = close_gaps_in_threshold(sauvola_result_denoised)

    blank, blank_2 = binarise_bead(len_image_x, len_image_y, sauvola_result_denoised)
    with open(text_directory, 'a') as text_file:
        print_log(f'        Image denoised and binarised;', file=text_file)

    # with image without background, we can remove substrate
    # to achieve this, we first need to compute the substrate line
    sum_blank_y = get_y_pixel_values(blank_2)

    sum_blank_y_filtered = low_pass_filter(np.linspace(0, len_image_y, len_image_y), sum_blank_y)
    sum_blank_y_derivative = np.gradient(sum_blank_y_filtered)

    # substrate_line_index = find_substrate_line_index(sum_blank_y, np.gradient(sum_blank_y))
    substrate_line_index = find_substrate_line_index(sum_blank_y_filtered, sum_blank_y_derivative)
    blank[substrate_line_index:, :] = 1

    with open(text_directory, 'a') as text_file:
        print_log(f'        Substrate found;', file=text_file)

    sum_blank_x = get_x_pixel_values(blank[:substrate_line_index])
    sum_blank_x_filtered = low_pass_filter(np.linspace(0, len_image_x, len_image_x), sum_blank_x)


    sum_blank_x_derivative = np.gradient(sum_blank_x_filtered)

    left_limit_1 = np.argmin(sum_blank_x_derivative)
    right_limit_1 = np.argmax(sum_blank_x_derivative)

    non_false_pixels_x = np.where(blank == 0)[1]
    non_false_pixels_y = np.where(blank == 0)[0]

    left_point_x = min(non_false_pixels_x)

    right_point_x = max(non_false_pixels_x)


    # invert values from 0-1
    x_values_without_substrate_inverted = [1 - value for value in sum_blank_x_filtered]

    # gets a first estimate to the middle of the line bead
    midway = np.argmin(sum_blank_x_filtered)
    # establishes a reference value that defines the width of the line bead
    reference_val = 1 - sum_blank_x_filtered[midway] * 0.05
    ind = midway

    # finds limits
    while sum_blank_x[ind] < reference_val and ind > 0:
        ind -= 1
    left_limit = ind

    ind = midway
    while sum_blank_x[ind] < reference_val and ind < len(sum_blank_x) - 1:
        ind += 1
    right_limit = ind


    if left_limit == 0 or right_limit > len_image_x -20:
        reference_val = 1 - sum_blank_x[midway] * 0.15
        ind = midway
        while sum_blank_x[ind] < reference_val and ind > 0:
            ind -= 1
        left_limit = ind

        ind = midway
        while sum_blank_x[ind] < reference_val and ind < len(
                sum_blank_x) - 1:
            ind += 1
        right_limit = ind

    if left_limit == 0 or right_limit > len_image_x -20:
        reference_val = 1 - sum_blank_x[midway] * 0.5
        ind = midway
        while sum_blank_x[ind] < reference_val and ind > 0:
            ind -= 1
        left_limit = ind

        ind = midway
        while sum_blank_x[ind] < reference_val and ind < len(
                sum_blank_x) - 1:
            ind += 1
        right_limit = ind

        with open(text_directory, 'a') as text_file:
            print('        Horizontal limits found;', file = text_file)

    blank[:, :left_limit] = 1
    blank[:, right_limit:] = 1

    # removing background
    image_gray_transparent = image_gray.copy()
    for pixel_x in range(0, len_image_x):
        for pixel_y in range(0, len_image_y):
            if blank[pixel_y, pixel_x] == True:
                image_gray_transparent[pixel_y, pixel_x] = 0

    # establishes two coordinates:
    #   one that underestimates the left timit;
    #   one that overestimates the right limit;
    # these coordinates are useful during the wettabilitty computation
    left_lim_adapted = int(round(0.8 * left_limit, 0))
    right_lim_adapted = int(round(1.2 * right_limit, 0))

    # crops two images necessary to determine the penetration limit
    cropped_image_gray = image_gray.copy()
    cropped_image_gray = crop_image(cropped_image_gray, (left_lim_adapted, right_lim_adapted), (substrate_line_index, len_image_y))
    cropped_image_blank = crop_image(blank, (left_limit, right_limit), (0, substrate_line_index))

    # extracts sum of y values, and approppriate derivative
    sum_cropped_y = get_y_pixel_values(cropped_image_gray)
    sum_cropped_y_filtered = low_pass_filter(np.linspace(0, len(sum_cropped_y), len(sum_cropped_y)), sum_cropped_y)
    sum_cropped_y_derivative = np.gradient(sum_cropped_y_filtered)

    # obtains penetration limit
    penetration_lim = substrate_line_index + np.argmax(sum_cropped_y_derivative)

    with open(text_directory, 'a') as text_file:
        print_log('        Penetration limit found;', file=text_file)

    """ Compute metrics """

    # Height
    (max_point_y, max_point_x) = get_line_bead_maximum(blank)

    bead_height_px = substrate_line_index - max_point_y
    bead_height_mm = get_height(max_point_y, substrate_line_index, px_2_mm)
    with open(text_directory, 'a') as text_file:
        print_log(f'        Bead Height: {round(bead_height_mm, 3)} mm', file=text_file)
    height_list.append(bead_height_mm)

    # Bead width
    bead_width = get_width(left_limit, right_limit, px_2_mm)
    with open(text_directory, 'a') as text_file:
        print_log(f'        Bead Width: {round(bead_width, 3)} mm', file=text_file)
    width_list.append(bead_width)

    # Penetration
    bead_penetration = get_height(substrate_line_index, penetration_lim, px_2_mm)
    with open(text_directory, 'a') as text_file:
        print_log(f'        Penetration: {round(bead_penetration, 3)} mm', file=text_file)
    penetration_list.append(bead_penetration)


    # Wettability Angles
    blank[substrate_line_index:, :] = blank_2[substrate_line_index:, :]
    for cri in [fraction_height_wettability]:   # this allows several wettabilities to be computed, according to
                                                # several height fractions; however, the

        left_slope, left_wettability_angle = get_left_wettability_angle(blank, left_lim_adapted, substrate_line_index, bead_height_px, cri, dy=2)
        with open(text_directory, 'a') as text_file:
            print_log(f'        Left wettability angle: {round(left_wettability_angle, 3)}°', file=text_file)
        left_wettability_angle_list.append(left_wettability_angle)

        right_slope, right_wettability_angle = get_right_wettability_angle(blank, right_lim_adapted, substrate_line_index, bead_height_px, cri, dy=2)
        with open(text_directory, 'a') as text_file:
            print_log(f'        Right wettability angle: {round(right_wettability_angle, 3)}°', file=text_file)
        right_wettability_angle_list.append(right_wettability_angle)

        # Area above substrate
        area_above_binarised_inverted = np.zeros((len_image_y, len_image_x))
        for pixel_x in range(0, len_image_x):
            for pixel_y in range(0, len_image_y):
                if blank[pixel_y, pixel_x] == 1:
                    area_above_binarised_inverted[pixel_y, pixel_x] = 0
                else:
                    area_above_binarised_inverted[pixel_y, pixel_x] = 1

        image_label = skimage.measure.label(area_above_binarised_inverted)
        props = skimage.measure.regionprops_table(image_label, properties=('centroid', 'area'))
        centroid_above_substrate = (props['centroid-1'][0], props['centroid-0'][0])
        area_above_substrate = props['area'][0] * px_2_mm ** 2
        with open(text_directory, 'a') as text_file:
            print_log(f'        Area above substrate: {round(area_above_substrate, 3)} mm ** 2', file=text_file)
        area_above_substrate_list.append(area_above_substrate)

    # Area below the substrate

    image_gray_cropped = image_gray.copy()
    image_gray_cropped = image_gray_cropped[substrate_line_index:penetration_lim, left_limit:right_limit]

    # performs color correction
    image_gray_corrected_log = skimage.exposure.adjust_gamma(image_gray_cropped, 3)
    image_gray_equalized = skimage.exposure.equalize_hist(image_gray_corrected_log)

    # gets sum of x and y pixel values, specifically from the dilution zone
    sum_dilution_zone_x = get_x_pixel_values(image_gray_cropped)
    sum_dilution_zone_y = get_y_pixel_values(image_gray_cropped)

    sum_dilution_zone_x_filtered = low_pass_filter(np.linspace(0, len(sum_dilution_zone_x), len(sum_dilution_zone_x)), sum_dilution_zone_x)
    dilution_zone = image_gray[substrate_line_index:penetration_lim, left_limit:right_limit].copy()

    dilution_zone_min = min(sum_dilution_zone_x_filtered)
    dilution_zone_max = max(sum_dilution_zone_x_filtered)

    # maps the sum of values to the image
    area_below_substrate_bi = np.full((len(sum_dilution_zone_y), len(sum_dilution_zone_x_filtered)), 1)
    for pixel_x in range(0, len(sum_dilution_zone_x_filtered)):
        pixel_y = int(round(- ((sum_dilution_zone_x_filtered[pixel_x] - dilution_zone_min) / (dilution_zone_max - dilution_zone_min)) * len(dilution_zone[:, 0]) + len(dilution_zone[:, 0]), 0))
        for pix_y in range(pixel_y, len(dilution_zone[:, 0])):
            area_below_substrate_bi[pix_y, pixel_x] = 0
    
    area_below_binarised_inverted = np.full((len_image_y, len_image_x), 0)
    area_below_binarised_inverted[substrate_line_index:penetration_lim, left_limit:right_limit] = area_below_substrate_bi
    image_label = skimage.measure.label(area_below_binarised_inverted[1:])

    props = skimage.measure.regionprops_table(image_label, properties=('centroid', 'area'))
    centroid_below_substrate = (props['centroid-1'][0], props['centroid-0'][0])
    area_below_substrate = props['area'][0] * px_2_mm ** 2
    with open(text_directory, 'a') as text_file:
        print_log(f'        Area below substrate:  {area_below_substrate} mm ** 2', file=text_file)
    area_below_substrate_list.append(area_below_substrate)

    # Dilution Proportion
    dilution_proportion = area_below_substrate / (area_below_substrate + area_above_substrate) * 100
    dilution_proportion_list.append(dilution_proportion)
    with open(text_directory, 'a') as text_file:
        print_log(f'        Dilution proportion: {round(dilution_proportion, 3)} %;', file=text_file)


    # Draws in image
    line_mask = np.full((len_image_y, len_image_x), 0)

    draw_line_from_slope(line_mask, substrate_line_index, left_limit, left_slope, 1)
    draw_line_from_slope(line_mask, substrate_line_index, right_limit, right_slope, 1)
    # draws penetration limit
    rr, cc = skimage.draw.line(penetration_lim, 0, penetration_lim, len_image_x - 1)
    line_mask[rr, cc] = 1
    # draws right limit
    rr, cc = skimage.draw.line(0, right_limit, len_image_y - 1, right_limit)
    line_mask[rr, cc] = 1
    # draws left limit
    rr, cc = skimage.draw.line(0, left_limit, len_image_y - 1, left_limit)
    line_mask[rr, cc] = 1
    # draws substrate line
    rr, cc = skimage.draw.line(substrate_line_index, 0, substrate_line_index, len_image_x - 1)
    line_mask[rr, cc] = 1
    # draws top of the line bead
    rr, cc = skimage.draw.line(max_point_y, 0, max_point_y, len_image_x - 1)
    line_mask[rr, cc] = 1

    counter += 1
    name = str(current_bead) + '-info.png'

    if save_image_collection == 1:
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(8, 8))
        ax = axes.ravel()
        ax[0].imshow(image_original, cmap='Greys')
        ax[0].set_title('Original image')

        ax[1].plot(sum_blank_y[::-1], np.linspace(0, len_image_y, len_image_y), 'b', label='sum_blank_y')
        ax[1].plot(sum_blank_y_filtered[::-1], np.linspace(0, len_image_y, len_image_y), 'r', label='filtered')
        ax[1].set_title('Sum of pixel\n values by row')
        ax[1].legend()

        ax[2].plot(sum_blank_y_derivative[::-1], np.linspace(0, len_image_y, len_image_y), 'b')
        ax[2].set_title('Derivaitve of pixel\n values by row')

        ax[3].plot(sum_blank_x, 'b', label='sum_blank_x')
        ax[3].plot([0, len_image_x], [reference_val, reference_val], 'r', label='Threshold')
        ax[3].set_title('Sum of pixel\n values by column')
        ax[3].legend()

        ax[4].imshow(sauvola_result_denoised)
        ax[4].set_title('Sauvola threshold+\n denoising')

        ax[5].imshow(area_below_binarised_inverted, 'Oranges', alpha=0.5)
        ax[5].imshow(area_above_binarised_inverted, 'Greens', alpha=0.5)
        circle_above = plt.Circle(centroid_above_substrate, round((len_image_x + len_image_y) / 100, 0), color='g')
        circle_below = plt.Circle(centroid_below_substrate, round((len_image_x + len_image_y) / 100, 0), color='r')
        ax[5].add_patch(circle_above)
        ax[5].add_patch(circle_below)
        ax[5].set_title('Areas above, below\n and centroids')

        plt.show
        plt.savefig(os.path.join(save_directory, name), dpi=300)
        plt.close(fig)


    if save_images == 1:
        name_2 = str(current_bead) + '-main.png'
        fig = plt.figure()
        plt.imshow(image_original, 'Greys')
        plt.imshow(line_mask, 'Blues', alpha=0.5)
        plt.imshow(area_below_binarised_inverted, 'Reds', alpha=0.3)
        plt.imshow(area_above_binarised_inverted, 'Greens', alpha=0.3)
        plt.show
        plt.savefig(os.path.join(save_directory, name_2), dpi=300)



    end_cycle_time = time.time()
    with open(text_directory, 'a') as text_file:
        print_log(f'    Bead  {current_bead}  analysed; ', file=text_file)
        print_log(f'    Time elapsed: {end_cycle_time - start_cycle_time}', file=text_file)


bead_data = {
    'name' : line_bead_name_list,
    'width' : width_list,
    'height' : height_list,
    'penetration' : penetration_list,
    'left angle' : left_wettability_angle_list,
    'right angle' : right_wettability_angle_list,
    'area above' : area_above_substrate_list,
    'area below' : area_below_substrate_list,
    'dilution' : dilution_proportion_list
}

if save_data == 1:
    with open(text_directory, 'a') as text_file:
        print_log(f'Saving database...', file=text_file)
    bead_dataframe = pd.DataFrame(bead_data, index=None)
    bead_dataframe.to_csv("Output.csv", index=False)
    with open(text_directory, 'a') as text_file:
        print_log(f'Database saved!', file=text_file)

end_time = time.time()
with open(text_directory, 'a') as text_file:
    print_log(f'Program concluded;', file=text_file)
    print_log(f'Total time elapsed: {end_time - start_time}', file=text_file)


